#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sub-package contains all necessary functions to compute the aperture
distribution and the (field) radiation pattern, and is ready to be used by the
least squares minimization.
"""

from .aperture import *
