#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This sub-package contains the geometrical properties of radio telescopes. There
are ready to use specific telescope structures and a
`~pyoof.telgeometry.block_manual` for different set ups.
"""

from .telgeometry import *
